@@ -2,41 +2,39 @@
 Using an IvyServer
 ------------------
 
-The following code is a typical example of use:
+The following code is a typical example of use::
 
-.. python::
     from ivy.ivy import IvyServer
-    
+
     class MyAgent(IvyServer):
-      def __init__(self, name):
-        IvyServer.__init__(self,'MyAgent')
-        self.name = name
+      def __init__(self, agent_name):
+        IvyServer.__init__(self,agent_name)
         self.start('127.255.255.255:2010')
         self.bind_msg(self.handle_hello, 'hello .*')
         self.bind_msg(self.handle_button, 'BTN ([a-fA-F0-9]+)')
-        
+
       def handle_hello(self, agent):
-        print '[Agent %s] GOT hello from %r'%(self.name, agent)
-      
+        print('[Agent %s] GOT hello from %r'%(self.agent_name, agent))
+
       def handle_button(self, agent, btn_id):
-        print '[Agent %s] GOT BTN button_id=%s from %r'%(self.name, btn_id, agent)
+        print('[Agent %s] GOT BTN button_id=%s from %r'%(self.agent_name, btn_id, agent))
         # let's answer!
         self.send_msg('BTN_ACK %s'%btn_id)
-    
+
     a=MyAgent('007')
 
 
 Implementation details
 ----------------------
 
-An Ivy client is made of several threads:
+An Ivy agent is made of several threads:
 
-  - an `IvyServer` instance
+  - an :class:`IvyServer` instance
 
   - a UDP server, launched by the Ivy server, listening to incoming UDP
     broadcast messages
 
-  - `IvyTimer` objects 
+  - :class:`IvyTimer` objects
 
 :group Messages types: BYE, ADD_REGEXP, MSG, ERROR, DEL_REGEXP, END_REGEXP,
   END_INIT, START_REGEXP, START_INIT, DIRECT_MSG, DIE
@@ -46,7 +44,7 @@ An Ivy client is made of several threads:
 :group Objects and functions related to logging: ivylogger, debug, log, warn,
   error, ivy_loghdlr, ivy_logformatter
 
-Copyright (c) 2005-2016 Sebastien Bigaret <sbigaret@users.sourceforge.net>
+Copyright (c) 2005-2021 Sebastien Bigaret <sebastien.bigaret@telecom-bretagne.eu>
 """
 import sys
 # The next line taken from https://pythonhosted.org/six/
@@ -87,7 +85,6 @@ ivy_logformatter = logging.Formatter('%(asctime)s %(levelname)s %(message)s')
 
 ivy_loghdlr.setFormatter(ivy_logformatter)
 ivylogger.addHandler(ivy_loghdlr)
-ivylogger.setLevel(logging.INFO)
 
 ##
 DEFAULT_IVYBUS = '127:2010'
@@ -227,7 +224,7 @@ def UDP_init_and_listen(broadcast_addr, port, socket_server):
         # Since we already have a client's name and id, lets register it
         # (this was previously done in IvyHandler's __init__() only)
         # but we want to check that we did not receive more than once a
-        # broadcast coming from the same 
+        # broadcast coming from the same
         new_client = socket_server._get_client(ip, port_number, new_socket,
                                                appid, appname)
         if new_client is None:
@@ -289,11 +286,42 @@ def decode_ivybus(ivybus=None):
     return broadcast, port
 
 
+def decode_MSG_params(params):
+    '''Implements the special treatment of parameters in text messages
+    (message type: MSG).  The purpose here is to make sure that, when
+    their last parameter is not ETX-terminated, they are processed the
+    same way as in the reference ivy-c library.
+    '''
+    MISSING_ETX = 'Received a misformatted message: last parameter is not ETX-terminated'
+    if ARG_END in params:  # there is at least one parameter
+        # All parameters are ETX-terminated: we remove the last ARG_END==ETX
+        # before  calling split(ARG_END)...
+        if params[-1] == ARG_END:
+            params = params[:-1]
+        else:
+            # ... However if the last ETX is missing, we pretend it was here
+            # so that we behave exactly as ivy-c in this case.
+            warn(MISSING_ETX)
+
+        params = params.split(ARG_END)
+
+    elif len(params)>0:
+        # One parameter was transmitted but it has no trailing ARG_END/ETX:
+        # let's conform to ivy-c behaviour and return it
+        warn(MISSING_ETX)
+        params = (params,)
+
+    return params
+
+
 def decode_msg(msg):
     """
+    Extracts from an ivybus message its message type, the conveyed
+    numerical identifier and its parameters.
+
+    :return: msg_type, numerical_id, parameters
+    :except IvyMalformedMessage: if the message's type or numerical identifier are not integers
 
-    :return: msg_id, numerical_id, parameters
-    :except IvyMalformedMessage:
     """
     try:
         msg_id, _msg = msg.split(' ', 1)
@@ -301,9 +329,11 @@ def decode_msg(msg):
         num_id, params = _msg.split(ARG_START, 1)
         num_id = int(num_id)
 
-        if ARG_END in params:
-            # there is an extra ARG_END after the last parameter and
-            # before the newline
+        if msg_id == MSG:
+            params = decode_MSG_params(params)
+
+        elif ARG_END in params:
+            # Remove the trailing ARG_END before calling split()
             params = params[:-1].split(ARG_END)
 
     except ValueError:
@@ -360,7 +390,7 @@ class IvyClient:
     Represents a client connected to the bus. Every callback methods
     registered by an agent receive an object of this type as their first
     parameter, so that they know which agent on the bus is the cause of the
-    event which triggered the callback.    
+    event which triggered the callback.
 
     An IvyClient is responsible for:
 
@@ -369,7 +399,7 @@ class IvyClient:
       - sending messages to the remote agent.
 
     It is **not** responsible for receiving messages from the client: another
-    object is in charge of that, namely an `IvyHandler` object.
+    object is in charge of that, namely an :class:`IvyHandler` object.
 
     The local IvyServer creates one IvyClient per agent on the bus.
 
@@ -377,11 +407,13 @@ class IvyClient:
     ---------
     See the discussion in `regexps`.
 
-    :group Protocol-related methods: start_init, end_init,
-      send_new_subscription, remove_subscription, wave_bye
-    :group Manipulating the remote agent's subscriptions:
-      add_regexp, remove_regexp
-    :group Sending messages: send_msg, send_direct_message, send_die_message
+    :Protocol-related methods: :py:func:`start_init`, :py:func:`end_init`,
+      :py:func:`send_new_subscription`, :py:func:`remove_subscription`,
+      :py:func:`wave_bye`
+    :Manipulating the remote agent's subscriptions:
+      :py:func:`add_regexp`, :py:func:`remove_regexp`
+    :Sending messages: :py:func:`send_msg`, :py:func:`send_direct_message`,
+      :py:func:`send_die_message`
 
     :ivar regexps: a dictionary mapping subscriptions' ids (as delivered by
       `add_regexp`) to the corresponding regular expressions. Precisely, it
@@ -429,7 +461,7 @@ class IvyClient:
         ``END_INIT``).  After this method is called, we expect to receive the
         initial subscriptions for that client (or none); the initialization
         process completes after `end_init` is called.
-        
+
         :except IvyIllegalStateError: if the method has already been called
           once
         """
@@ -442,7 +474,7 @@ class IvyClient:
     def end_init(self):
         """
         Should be called when the initialization process ends.
-        
+
         :except IvyIllegalStateError: if the method has already been called
           (and ``self.status`` has already been set to ``INITIALIZED``)
         """
@@ -454,7 +486,7 @@ class IvyClient:
     def add_regexp(self, regexp_id, regexp):
         """
         :except IvyIllegalStateError: if the client has not been fully
-          initialized yet (see `start_init`)
+          initialized yet (see :py:func:`start_init`)
         """
         if self.status not in (INITIALIZATION_IN_PROGRESS, INITIALIZED):
             # initialization has not begun
@@ -471,10 +503,10 @@ class IvyClient:
     def remove_regexp(self, regexp_id):
         """
         Removes a regexp
-        
+
         :return: the regexp that has been removed
         :except IvyIllegalStateError: if the client has not been fully
-          initialized yet (see `start_init`)
+          initialized yet (see :py:func:`start_init`)
         :except KeyError: if no such subscription exists
         """
         if self.status not in (INITIALIZATION_IN_PROGRESS, INITIALIZED):
@@ -501,11 +533,11 @@ class IvyClient:
         """
         Sends the message to the client.  The client will receive one message
         for each of its subscriptions (regexps) that the message matches.
-        
+
         :return: ``True`` if the message was actually sent to the client, that
           is: if there is one or more regexps matching the message in the
           client's subscriptions; returns ``False`` otherwise.
-        
+
         """
         if self.status != INITIALIZED:
             return
@@ -546,7 +578,7 @@ class IvyClient:
         design, supported by python, but if you want to inter-operate with
         applications using the standard Ivy API the message you send *must* be
         a string. See in particular in ``ivy.h``::
-        
+
           typedef void (*MsgDirectCallback)( IvyClientPtr app, void *user_data, int id, char *msg ) ;
 
         """
@@ -676,7 +708,7 @@ class IvyClient:
 
         .. note:: ``ivysocket.c:SocketSendRaw()`` also ignores error, simply
            logging them.
-        
+
         """
         try:
             self.socket.send(encode_message(msg_type, *params))
@@ -715,13 +747,13 @@ class IvyServer(socketserver.ThreadingTCPServer):
     All public methods (not starting with an underscore ``_``) are
     MT-safe
 
-    :group Communication on the ivybus: start, send_msg, send_direct_message,
-      send_ready_message, handle_msg, stop
+    :group Communication on the ivybus: :py:func:`start`, :py:func:`send_msg`, :py:func:`send_direct_message`,
+      :py:func:`send_ready_message`, :py:func:`handle_msg`, :py:func:`stop`
 
-    :group Inspecting the ivybus: get_clients, _get_client, get_client_with_name
+    :group Inspecting the ivybus: :py:func:`get_clients`, :py:func:`_get_client`, :py:func:`get_client_with_name`
 
-    :group Our own subscriptions: get_subscriptions, bind_msg, unbind_msg,
-      _add_subscription, _remove_subscription, _get_fct_for_subscription
+    :group Our own subscriptions: :py:func:`get_subscriptions`, :py:func:`bind_msg`, :py:func:`unbind_msg`,
+      :py:func:`_add_subscription`, :py:func:`_remove_subscription`, :py:func:`_get_fct_for_subscription`
 
     """
     # Impl. note: acquiring/releasing the global lock in methods
@@ -738,9 +770,9 @@ class IvyServer(socketserver.ThreadingTCPServer):
         newly created instances to connect to the corresponding Ivy bus and to
         start communicating with other applications.
 
-        MT-safety: both functions `app_callback` and `die_callback` must be
-        prepared to be called concurrently
-        
+        MT-safety: both functions :py:func:`app_callback` and
+        :py:func:`die_callback` must be prepared to be called concurrently
+
         :Parameters:
           - `agent_name`: the client's agent name
           - `ready_msg`: a message to send to clients when they connect
@@ -751,7 +783,7 @@ class IvyServer(socketserver.ThreadingTCPServer):
           - `die_callback`: called when the IvyServer receives a DIE message
           - `usesDaemons`: see above.
 
-        :see: `bind_msg()`, `start()`
+        .. seealso:: :py:func:`bind_msg()`, :py:func:`start()`
         """
         self._thread = None
 
@@ -787,16 +819,43 @@ class IvyServer(socketserver.ThreadingTCPServer):
 
         self.agent_id = agent_name + time.strftime('%Y%m%d%H%M%S') + '%05i' % random.randint(0, 99999) + str(self.port)
 
+    @staticmethod
+    def run_callback(callback, callback_description,
+                     agent, *args, **kw):
+        """Runs a callback, catching any exception it may raise.
+
+        :Parameters:
+         - `callback`: the function to be called
+         - `callback_description`: the description to use in the error message,
+           when an exception is raised.
+         - `agent`: the :py:class:`IvyClient` triggering the callback, which
+           is passed as the first argument to the callback
+         - `on_exc`: the returned value in case an exception was raised by
+           the callback.
+         - All other arguments are passed as-is to the callback.
+
+        :return: the value returned by the callback, or `exc_none` if
+          an exception was raised
+
+        """
+        on_exc = kw.get('on_exc', None)
+        try:
+            return callback(agent, *args, **kw)
+        except:
+            error(callback_description + ': exception raised',
+                  exc_info=sys.exc_info())
+            return on_exc
+
     def serve_forever(self):
         """
-        Handle requests (calling `handle_request()`) until doomsday... or
-        until `stop()` is called.
+        Handle requests (calling :py:func:`handle_request()`) until doomsday... or
+        until :py:func:`stop()` is called.
 
         This method is registered as the target method for the thread.
         It is also responsible for launching the UDP server in a separate
-        thread, see `UDP_init_and_listen` for details.
-        
-        You should not need to call this method, use `start` instead.
+        thread, see :py:func:`UDP_init_and_listen` for details.
+
+        You should not need to call this method, use :py:func:`start` instead.
         """
         broadcast, port = decode_ivybus(self.ivybus)
         l = lambda server=self: UDP_init_and_listen(broadcast, port, server)
@@ -813,7 +872,7 @@ class IvyServer(socketserver.ThreadingTCPServer):
     def start(self, ivybus=None):
         """
         Binds the server to the ivybus. The server remains connected until
-        `stop` is called, or until it receives and accepts a 'die' message.
+        :py:func:`stop` is called, or until it receives and accepts a 'die' message.
 
         :except IvyIllegalStateError: if the server has already been
           started
@@ -833,7 +892,7 @@ class IvyServer(socketserver.ThreadingTCPServer):
     def stop(self):
         """
         Disconnects the server from the ivybus. It also sets the
-        `server_termination` event.
+        :py:func:`server_termination` event.
 
         :except IvyIllegalStateError: if the server is not running started
         """
@@ -857,7 +916,7 @@ class IvyServer(socketserver.ThreadingTCPServer):
     def isAlive(self):
         if self._thread is None:
             return False
-        return self._thread.isAlive()
+        return self._thread.is_alive()
 
     def get_clients(self):
         """
@@ -881,7 +940,7 @@ class IvyServer(socketserver.ThreadingTCPServer):
         same id is already registered; if it exists, the method exits by
         returning None.
 
-        You should not need to call this, use `get_client_with_name` instead
+        You should not need to call this, use :py:func:`get_client_with_name` instead
         """
         self._global_lock.acquire()
         try:
@@ -915,11 +974,16 @@ class IvyServer(socketserver.ThreadingTCPServer):
         TODO: maybe add a flag (while connecting) on the client, that would prevent sending msg. etc..
         as CNX. not confirmed
         """
-        self.app_callback(client, IvyApplicationConnected)
+        self.run_callback(self.app_callback,
+                          'application callback (connection)',
+                          client, IvyApplicationConnected)
 
     def handle_die_message(self, msg_id, from_client=None):
         """  """
-        should_die = (self.die_callback(from_client, msg_id) != IVY_SHOULD_NOT_DIE)
+        should_die = self.run_callback(self.die_callback, 'die callback',
+                                       from_client, msg_id)
+        should_die = should_die != IVY_SHOULD_NOT_DIE
+
         log('Received a die msg from: %s with id: %s -- should die=%s',
             from_client or '<unknown>', msg_id, should_die)
         if should_die:
@@ -935,7 +999,9 @@ class IvyServer(socketserver.ThreadingTCPServer):
         """
         log('Received a direct msg from: %s with id: %s -- %s',
             client or '<unknown>', num_id, msg)
-        self.direct_callback(client, num_id, msg)
+        description = "direct message callback: num_id:%s msg:%s"%(num_id, msg)
+        self.run_callback(self.direct_callback, description,
+                          client, num_id, msg)
 
     def handle_regexp_change(self, client, event, num_id, regexp):
         """
@@ -948,12 +1014,18 @@ class IvyServer(socketserver.ThreadingTCPServer):
             event = IvyRegexpAdded
         else:
             event = IvyRegexpRemoved
-        self.regexp_change_callback(client, event, num_id, regexp)
+        description = ( "regexp change callback ("
+                        + (ADD_REGEXP and 'add' or 'remove')
+                        + "): num_id: %s regexp:%s"%(num_id, regexp) )
+        self.run_callback(self.regexp_change_callback, description,
+                          client, event, num_id, regexp)
 
     def handle_pong(self, client, delta):
         """
         """
-        self.pong_callback(client, delta)
+        log('Received a pong reply from: %s with delta: %s',
+            client or '<unknown>', delta)
+        self.run_callback(self.pong_callback, 'pong callback', client, delta)
 
     def remove_client(self, ip, port, trigger_application_callback=True):
         """
@@ -975,8 +1047,9 @@ class IvyServer(socketserver.ThreadingTCPServer):
             debug('Removing client %r', removed_client)
             del self._clients[removed_client]
             if trigger_application_callback:
-                self.app_callback(removed_client, IvyApplicationDisconnected)
-
+                self.run_callback(self.app_callback,
+                                  'application callback (disconnection)',
+                                  removed_client, IvyApplicationDisconnected)
             return removed_client
         finally:
             self._global_lock.release()
@@ -1085,12 +1158,17 @@ class IvyServer(socketserver.ThreadingTCPServer):
         self._global_lock.acquire()
         try:
             try:
-                return self._get_fct_for_subscription(int(idx))(client, *params)
+                regexp, callback = self._subscriptions[int(idx)]
             except KeyError:
                 # it is possible that we receive a message for a regexp that
                 # was subscribed then unregistered
                 warn('Asked to handle an unknown subscription: id:%r params: %r'
                      ' --ignoring', idx, params)
+                return
+            self.run_callback(callback,
+                              'callback for subscription %i (%s)'%(idx, regexp),
+                              client,
+                              *params)
         finally:
             self._global_lock.release()
 
@@ -1134,7 +1212,7 @@ class IvyServer(socketserver.ThreadingTCPServer):
           - `regexp`: (string) a regular expression
 
         :return: the binding's id, which can be used to unregister the binding
-          with `unbind_msg()`
+          with :py:func:`unbind_msg()`
         """
         self._global_lock.acquire()
         idx = self._add_subscription(regexp, on_msg_fct)
@@ -1149,7 +1227,7 @@ class IvyServer(socketserver.ThreadingTCPServer):
         """
         Unbinds a subscription
 
-        :param num_id: the binding's id, as returned by `bind_msg()`
+        :param num_id: the binding's id, as returned by :py:func:`bind_msg()`
 
         :return: the regexp corresponding to the unsubscribed binding
         :except KeyError: if no such subscription can be found
@@ -1273,19 +1351,19 @@ class IvyHandler(socketserver.StreamRequestHandler):
 
     def process_ivymessage(self, msg, client):
         """
-        Examines the message (after passing it through the `decode_msg()`
+        Examines the message (after passing it through the :py:func:`decode_msg()`
         filter) and takes the appropriate actions depending on the message
         types.  Please refer to the document `The Ivy Architecture and
-        Protocol <http://www.tls.cena.fr/products/ivy/documentation>`_ and to
+        Protocol <http://www.eei.cena.fr/products/ivy/documentation>`_ and to
         the python code for further details.
-        
-        
+
+
         :Parameters:
           - `msg`: (should not include a newline at end)
 
         :return: ``False`` if the connection should be terminated, ``True``
           otherwise
-        
+
         """
         # cf. static void Receive() in ivy.c
         try:
@@ -1357,7 +1435,7 @@ class IvyHandler(socketserver.StreamRequestHandler):
                 self.server.handle_die_message(num_id, client)
 
             elif msg_id == DIRECT_MSG:
-                # num_id: not meaningful. 
+                # num_id: not meaningful.
                 log('Client %r sent us a direct msg num_id:%s msg:%r',
                     client, num_id, params)
                 self.server.handle_direct_msg(client, num_id, params)
@@ -1386,47 +1464,49 @@ class IvyHandler(socketserver.StreamRequestHandler):
 class IvyTimer(threading.Thread):
     """
     An IvyTimer object is responsible for calling a function regularly.  It is
-    bound to an IvyServer and stops when its server stops.
+    bound to an :py:class:`IvyServer` and stops when its server stops.
 
-    Interacting with a timer object
-    -------------------------------
+    Interacting with a timer object:
+      - Each timer gets a unique id, stored in the attribute ``id``. Note that
+        a dead timer's id can be reassigned to another one (a dead timer is a
+        timer that has been stopped)
 
-    - Each timer gets a unique id, stored in the attribute ``id``. Note that
-      a dead timer's id can be reassigned to another one (a dead timer is a
-      timer that has been stopped)
+      - To start a timer, simply call its method :py:func:`start`
 
-    - To start a timer, simply call its method ``start()``
-    
-    - To modify a timer's delay: simply assign ``timer.delay``, the
-      modification will be taken into account after the next tick. The delay
-      should be given in milliseconds.
+      - To modify a timer's delay: simply assign :py:attr:`delay`, the
+        modification will be taken into account after the next tick. The delay
+        should be given in milliseconds.
 
-    - to stop a timer, assign ``timer.abort`` to ``True``, the timer will stop
-      at the next tick (the callback won't be called)
+      - to stop a timer, assign :py:attr:`abort` to ``True``, the timer will stop
+        at the next tick (the callback won't be called)
 
-    MT-safety
-    ---------
-    **Please note:** ``start()`` starts a new thread; if the same function is
-    given as the callback to different timers, that function should be
-    prepared to be called concurrently.  Specifically, if the callback
-    accesses shared variables, they should be protected against concurrency
-    problems (using locks e.g.).
+    MT-safety:
+      **Please note:** ``start()`` starts a new thread; if the same function is
+      given as the callback to different timers, that function should be
+      prepared to be called concurrently.  Specifically, if the callback
+      accesses shared variables, they should be protected against concurrency
+      problems (using locks e.g.).
+
+
+    .. py:attribute:: abort
+
+    Stops the timer at the next tick.
     """
 
     def __init__(self, server, nbticks, delay, callback):
         """
         Creates a new timer.  After creation, call the timer's ``start()``
         method to activate it.
-        
+
         :Parameters:
-          - `server`: the `IvyServer` related to this timer --when the server
+          - `server`: the :py:class:`IvyServer` related to this timer --when the server
             stops, so does the timer.
           - `nbticks`: the number of repetition to make. ``0`` (zero) means:
             endless loop
           - `delay`: the delay, in milliseconds, between two ticks
           - `callback`: a function called at each tick. This function is
             called with one parameter, the timer itself
-            
+
         """
         threading.Thread.__init__(self)
         self.server = server
@@ -1434,6 +1514,7 @@ class IvyTimer(threading.Thread):
         self.delay = delay       # milliseconds
         self.callback = callback
         self.abort = False
+        #: The timer unique id
         self.id = id(self)
         self.setDaemon(server.usesDaemons)
 
